/*
 * Copyright (c) 2018, 2019, Oracle and/or its affiliates. All rights reserved.
 * DO NOT ALTER OR REMOVE COPYRIGHT NOTICES OR THIS FILE HEADER.
 *
 * This code is free software; you can redistribute it and/or modify it
 * under the terms of the GNU General Public License version 2 only, as
 * published by the Free Software Foundation.
 *
 * This code is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
 * FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
 * version 2 for more details (a copy is included in the LICENSE file that
 * accompanied this code).
 *
 * You should have received a copy of the GNU General Public License version
 * 2 along with this work; if not, write to the Free Software Foundation,
 * Inc., 51 Franklin St, Fifth Floor, Boston, MA 02110-1301 USA.
 *
 * Please contact Oracle, 500 Oracle Parkway, Redwood Shores, CA 94065 USA
 * or visit www.oracle.com if you need additional information or have any
 * questions.
 *
 */

#ifndef SHARE_MEMORY_METASPACE_ABSTRACTPOOL_HPP
#define SHARE_MEMORY_METASPACE_ABSTRACTPOOL_HPP

#include "memory/allocation.hpp"
#include "utilities/debug.hpp"
#include "utilities/globalDefinitions.hpp"

namespace metaspace {

// A simple helper class;
//
// holds a linear array of elements of type E;
//
// array lives in C heap and expands automatically;
//
// free elements can be returned and are kept in a freelist;
//
// elements can be retrieved by their index (index type I).
//
// E must be of at least sizeof(I), and I must be an integral size.
//
template <class E, class I, I initial_size, I size_increase, I max_size>
class AbstractPool {
public:

  typedef I size_type_t;

  static const I invalid_idx = (I)0;

private:

  E* _arr;
  size_type_t _capacity;
  size_type_t _used;
  I _freelist;

  const char* const _name;

  // Number of elements in the freelist
  DEBUG_ONLY(size_type_t _num_in_freelist;)

  // Enlarge internal array if needed. Zero out new portion of array.
  void enlarge_to(size_type_t new_len) {
    assert(new_len > _capacity && new_len <= max_size, "Sanity");
    const size_t new_size_bytes = new_len * sizeof(E);
    E* p = (E*) os::realloc(_arr, new_size_bytes, mtInternal);
    if (p == NULL) {
      vm_exit_out_of_memory(new_size_bytes, OOM_MALLOC_ERROR, "Pool %s: not enough space", _name);
    }
    _arr = p; _capacity = new_len;
  }

  //// freelist handling ////

  struct free_elem_t {
    I next_idx;
  };

  STATIC_ASSERT(sizeof(E) >= sizeof(free_elem_t));
  STATIC_ASSERT(initial_size > 0);
  STATIC_ASSERT(size_increase > 0);
  STATIC_ASSERT(max_size >= size_increase + initial_size);

  E* take_from_freelist() {
    I idx = _freelist;
    if (invalid_idx == idx) {
      return NULL;
    }
    E* p = elem_at_index(idx);
    _freelist = ((free_elem_t*)p)->next_idx;
    assert(_num_in_freelist > 0, "counter underflow");
    DEBUG_ONLY(_num_in_freelist --;)
    return p;
  }

  void add_to_freelist(E* p) {
    ((free_elem_t*)p)->next_idx = _freelist;
    _freelist = index_for_elem(p);
    DEBUG_ONLY(_num_in_freelist ++;)
  }


#ifdef ASSERT
  void check_elem(E* p) { assert(is_valid_elem(p), "invalid pointer"); }
  void check_idx(I i) { assert(i <= _used, "invalid index"); }
#endif

public:

  AbstractPool(const char* name)
    : _arr(NULL), _capacity(0), _used(0), _name(name), _freelist(NULL) {}
  ~AbstractPool() { os::free(_arr); }

  I index_for_elem(const E* p) const {
    DEBUG_ONLY(check_elem(p));
    return p - _arr;
  }

  E* elem_at_index(I i) const {
    DEBUG_ONLY(check_idx(i));
    return _arr + i;
  }

  // Allocate a new element. Enlarge internal array if needed.
  // Will return NULL if max_size is reached.
  E* allocate_element() {

    E* p = take_from_freelist();
    if (p != NULL) {
      return p;
    }

    if (_capacity == _used) {
      size_type_t new_len = _capacity == 0 ? initial_size : _capacity + size_increase;
      if (new_len > max_size) {
        new_len = max_size;
        if (new_len <= _capacity) {
          return NULL;
        }
      }
      enlarge_to(new_len);
    }

    // Avoid handing out anything at index 0.
    // This allows callers to use "I index == 0" as "invalid ref".
    if (_used == 0) {
      _used ++;
    }

    E* p = _arr[_used ++];
    return p;

  }

  void return_element(E* p) {
    DEBUG_ONLY(check_elem(p));
    add_to_freelist(p);
  }

  // Returns true if p was allocated from this pool.
  bool is_valid_elem(const E* p) const { return p >= _arr && p < _arr + _used; }

  bool is_valid_index(I idx) const     { return idx < _used; }

  // Returns number of allocated elements (including those returned to free list)
  size_type_t used() const { return _used; }

  // Returns number of elements in free list
  size_type_t free() const { return _num_in_freelist; }

  // Returns size of memory used.
  size_t memory_footprint() const { return _capacity * sizeof(E); }

};


} // namespace metaspace

#endif // SHARE_MEMORY_METASPACE_ABSTRACTPOOL_HPP
