/*
 * Copyright (c) 2018, 2019, Oracle and/or its affiliates. All rights reserved.
 * Copyright (c) 2018 SAP SE. All rights reserved.
 * DO NOT ALTER OR REMOVE COPYRIGHT NOTICES OR THIS FILE HEADER.
 *
 * This code is free software; you can redistribute it and/or modify it
 * under the terms of the GNU General Public License version 2 only, as
 * published by the Free Software Foundation.
 *
 * This code is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
 * FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
 * version 2 for more details (a copy is included in the LICENSE file that
 * accompanied this code).
 *
 * You should have received a copy of the GNU General Public License version
 * 2 along with this work; if not, write to the Free Software Foundation,
 * Inc., 51 Franklin St, Fifth Floor, Boston, MA 02110-1301 USA.
 *
 * Please contact Oracle, 500 Oracle Parkway, Redwood Shores, CA 94065 USA
 * or visit www.oracle.com if you need additional information or have any
 * questions.
 *
 */

#ifndef SHARE_MEMORY_METASPACE_METASPACESTATISTICS_HPP
#define SHARE_MEMORY_METASPACE_METASPACESTATISTICS_HPP

#include "memory/metaspace.hpp"             // for MetadataType enum
#include "memory/metaspace/chunkLevel.hpp"
#include "utilities/globalDefinitions.hpp"

class outputStream;

namespace metaspace {

// Contains statistics for a number of free chunks.
struct FreeChunksStatistics {

  uintx num;         // Number of chunks
  size_t word_size;  // Total capacity, in words

  void reset() { num = 0; word_size = 0; }

  void add(uintx n, size_t s) {
    num += n; word_size += s;
  }

  void add(const FreeChunksStatistics& other) {
    num += other.num; word_size += other.word_size;
  }

  void print_on(outputStream* st, size_t scale) const;

}; // end: FreeChunksStatistics


// Contains statistics for a ChunkManager.
struct ChunkManagerStatistics {

  FreeChunksStatistics chunk_stats[chklvl::NUM_CHUNK_LEVELS];

  void reset();
  size_t total_capacity() const;

  void print_on(outputStream* st, size_t scale) const;

}; // ChunkManagerStatistics

// Contains statistics for a number of chunks in use.
// Each chunk has a used and free portion; however, there are current chunks (serving
// potential future metaspace allocations) and non-current chunks. Unused portion of the
// former is counted as free, unused portion of the latter counts as waste.
struct UsedChunksStatistics {

  uintx num;     // Number of chunks
  size_t cap;    // Total capacity in words.
  size_t used;   // Total used area, in words
  size_t free;   // Total free area (unused portions of current chunks), in words
  size_t waste;  // Total waste area (unused portions of non-current chunks), in words

  void reset() {
    num = 0;
    cap = used = free = waste = 0;
  }

  void add(const UsedChunksStatistics& other) {
    num = other.num; cap = other.cap; used = other.used; free = other.free; waste = other.waste;
  }

  void print_on(outputStream* st, size_t scale) const;

  DEBUG_ONLY(void check_sanity() const;)

}; // UsedChunksStatistics

// Class containing statistics for one or more space managers.
struct SpaceManagerStatistics {

  UsedChunksStatistics chunk_stats[chklvl::NUM_CHUNK_LEVELS];
  uintx free_blocks_num;
  size_t free_blocks_word_size;

  void reset();

  void add_free_blocks_info(uintx num, size_t cap) {
    free_blocks_num += num; free_blocks_word_size += cap;
  }

  // Returns total chunk statistics over all chunk levels.
  UsedChunksStatistics totals() const;

  void add(const SpaceManagerStatistics& other);

  void print_on(outputStream* st, size_t scale,  bool detailed) const;

}; // SpaceManagerStatistics

struct ClassLoaderMetaspaceStatistics {

  SpaceManagerStatistics sm_stats[Metaspace::MetadataTypeCount];

  void reset() {
    sm_stats[Metaspace::ClassType].reset();
    sm_stats[Metaspace::NonClassType].reset();
  }

  void add(const ClassLoaderMetaspaceStatistics& other) {
    sm_stats[Metaspace::ClassType].add(other.sm_stats[Metaspace::ClassType]);
    sm_stats[Metaspace::NonClassType].add(other.sm_stats[Metaspace::NonClassType]);
  }

  // Returns total space manager statistics for both class and non-class metaspace
  SpaceManagerStatistics totals() const;

  void print_on(outputStream* st, size_t scale, bool detailed) const;

}; // ClassLoaderMetaspaceStatistics

} // namespace metaspace

#endif // SHARE_MEMORY_METASPACE_METASPACESTATISTICS_HPP
